@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016, Intel Corporation
+ * Copyright (c) 2014-2018, Intel Corporation
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -29,7 +29,9 @@
 #include "ptunit.h"
 
 #include "pt_packet_decoder.h"
+#include "pt_query_decoder.h"
 #include "pt_encoder.h"
+#include "pt_opcodes.h"
 
 #include "intel-pt.h"
 
@@ -393,6 +395,75 @@ static struct ptunit_result mnt(struct packet_fixture *pfix)
 	return ptu_passed();
 }
 
+static struct ptunit_result exstop(struct packet_fixture *pfix, int ip)
+{
+	pfix->packet[0].type = ppt_exstop;
+	pfix->packet[0].payload.exstop.ip = ip ? 1 : 0;
+
+	ptu_test(pfix_test, pfix);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result mwait(struct packet_fixture *pfix)
+{
+	pfix->packet[0].type = ppt_mwait;
+	pfix->packet[0].payload.mwait.hints = 0xc;
+	pfix->packet[0].payload.mwait.ext = 0x1;
+
+	ptu_test(pfix_test, pfix);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result pwre(struct packet_fixture *pfix)
+{
+	pfix->packet[0].type = ppt_pwre;
+	pfix->packet[0].payload.pwre.state = 0x0;
+	pfix->packet[0].payload.pwre.sub_state = 0x3;
+	pfix->packet[0].payload.pwre.hw = 1;
+
+	ptu_test(pfix_test, pfix);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result pwrx(struct packet_fixture *pfix)
+{
+	pfix->packet[0].type = ppt_pwrx;
+	pfix->packet[0].payload.pwrx.last = 0x3;
+	pfix->packet[0].payload.pwrx.deepest = 0xa;
+	pfix->packet[0].payload.pwrx.store = 1;
+
+	ptu_test(pfix_test, pfix);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result ptw(struct packet_fixture *pfix, uint8_t plc,
+				int ip)
+{
+	uint64_t pl, mask;
+	int size;
+
+	size = pt_ptw_size(plc);
+	ptu_int_gt(size, 0);
+
+	pl = 0x1234567890abcdefull;
+
+	ptu_uint_le((size_t) size, sizeof(mask));
+	mask = ~0ull >> ((sizeof(mask) - (size_t) size) * 8);
+
+	pfix->packet[0].type = ppt_ptw;
+	pfix->packet[0].payload.ptw.payload = pl & mask;
+	pfix->packet[0].payload.ptw.plc = plc;
+	pfix->packet[0].payload.ptw.ip = ip ? 1 : 0;
+
+	ptu_test(pfix_test, pfix);
+
+	return ptu_passed();
+}
+
 static struct ptunit_result cutoff(struct packet_fixture *pfix,
 				   enum pt_packet_type type)
 {
@@ -546,6 +617,13 @@ int main(int argc, char **argv)
 	ptu_run_f(suite, cyc, pfix);
 	ptu_run_f(suite, vmcs, pfix);
 	ptu_run_f(suite, mnt, pfix);
+	ptu_run_fp(suite, exstop, pfix, 0);
+	ptu_run_fp(suite, exstop, pfix, 1);
+	ptu_run_f(suite, mwait, pfix);
+	ptu_run_f(suite, pwre, pfix);
+	ptu_run_f(suite, pwrx, pfix);
+	ptu_run_fp(suite, ptw, pfix, 0, 1);
+	ptu_run_fp(suite, ptw, pfix, 1, 0);
 
 	ptu_run_fp(suite, cutoff, pfix, ppt_psb);
 	ptu_run_fp(suite, cutoff_ip, pfix, ppt_tip);
@@ -564,9 +642,13 @@ int main(int argc, char **argv)
 	ptu_run_fp(suite, cutoff_mode, pfix, pt_mol_tsx);
 	ptu_run_fp(suite, cutoff, pfix, ppt_vmcs);
 	ptu_run_fp(suite, cutoff, pfix, ppt_mnt);
+	ptu_run_fp(suite, cutoff, pfix, ppt_exstop);
+	ptu_run_fp(suite, cutoff, pfix, ppt_mwait);
+	ptu_run_fp(suite, cutoff, pfix, ppt_pwre);
+	ptu_run_fp(suite, cutoff, pfix, ppt_pwrx);
+	ptu_run_fp(suite, cutoff, pfix, ppt_ptw);
 
-	ptunit_report(&suite);
-	return suite.nr_fails;
+	return ptunit_report(&suite);
 }
 
 
@@ -739,3 +821,39 @@ int pt_qry_decode_mnt(struct pt_query_decoder *d)
 
 	return -pte_internal;
 }
+int pt_qry_header_mnt(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
+int pt_qry_decode_exstop(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
+int pt_qry_decode_mwait(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
+int pt_qry_decode_pwre(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
+int pt_qry_decode_pwrx(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
+int pt_qry_decode_ptw(struct pt_query_decoder *d)
+{
+	(void) d;
+
+	return -pte_internal;
+}
