@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2016, Intel Corporation
+ * Copyright (c) 2013-2018, Intel Corporation
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -32,6 +32,7 @@
 #include "pt_decoder_function.h"
 #include "pt_query_decoder.h"
 #include "pt_encoder.h"
+#include "pt_opcodes.h"
 
 
 /* A query testing fixture. */
@@ -59,16 +60,16 @@ struct ptu_decoder_fixture {
 };
 
 /* An invalid address. */
-static const uint64_t pt_dfix_bad_ip = (1ull << 62) - 1;
+static const uint64_t pt_dfix_bad_ip = (1ull << 62) - 1ull;
 
 /* A sign-extended address. */
 static const uint64_t pt_dfix_sext_ip = 0xffffff00ff00ff00ull;
 
 /* The highest possible address. */
-static const uint64_t pt_dfix_max_ip = (1ull << 47) - 1;
+static const uint64_t pt_dfix_max_ip = (1ull << 47) - 1ull;
 
 /* The highest possible cr3 value. */
-static const uint64_t pt_dfix_max_cr3 = ((1ull << 47) - 1) & ~0x1f;
+static const uint64_t pt_dfix_max_cr3 = ((1ull << 47) - 1ull) & ~0x1full;
 
 /* Synchronize the decoder at the beginning of the trace stream, avoiding the
  * initial PSB header.
@@ -138,6 +139,306 @@ static struct ptunit_result event_not_synced(struct ptu_decoder_fixture *dfix)
 	return ptu_passed();
 }
 
+static struct ptunit_result sync_backward(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	uint64_t sync[3], offset, ip;
+	int errcode;
+
+	/* Check that we can use repeated pt_qry_sync_backward() to iterate over
+	 * synchronization points in backwards order.
+	 */
+
+	errcode = pt_enc_get_offset(encoder, &sync[0]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[1]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[2]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	/* Synchronize repeatedly and check that we reach each PSB in the
+	 * correct order.
+	 */
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[2]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[1]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[0]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_eq(errcode, -pte_eos);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result
+sync_backward_empty_end(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	uint64_t sync[3], offset, ip;
+	int errcode;
+
+	/* Check that we can use repeated pt_qry_sync_backward() to iterate over
+	 * synchronization points in backwards order.
+	 *
+	 * There's an empty PSB+ at the end.  We skip it.
+	 */
+
+	errcode = pt_enc_get_offset(encoder, &sync[0]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[1]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[2]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_psbend(encoder);
+
+	/* Synchronize repeatedly and check that we reach each PSB in the
+	 * correct order.
+	 */
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[1]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[0]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_eq(errcode, -pte_eos);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result
+sync_backward_empty_mid(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	uint64_t sync[3], offset, ip;
+	int errcode;
+
+	/* Check that we can use repeated pt_qry_sync_backward() to iterate over
+	 * synchronization points in backwards order.
+	 *
+	 * There's an empty PSB+ in the middle.  We skip it.
+	 */
+
+	errcode = pt_enc_get_offset(encoder, &sync[0]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[1]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[2]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	/* Synchronize repeatedly and check that we reach each PSB in the
+	 * correct order.
+	 */
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[2]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[0]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_eq(errcode, -pte_eos);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result
+sync_backward_empty_begin(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	uint64_t sync[3], offset, ip;
+	int errcode;
+
+	/* Check that we can use repeated pt_qry_sync_backward() to iterate over
+	 * synchronization points in backwards order.
+	 *
+	 * There's an empty PSB+ at the beginning.  We skip it.
+	 */
+
+	errcode = pt_enc_get_offset(encoder, &sync[0]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[1]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[2]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	/* Synchronize repeatedly and check that we reach each PSB in the
+	 * correct order.
+	 */
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[2]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[1]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_eq(errcode, -pte_eos);
+
+	return ptu_passed();
+}
+
+static struct ptunit_result
+decode_sync_backward(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	struct pt_event event;
+	uint64_t sync[2], offset, ip;
+	int errcode;
+
+	/* Check that we can use sync_backward to re-sync at the current trace
+	 * segment as well as to find the previous trace segment.
+	 */
+
+	errcode = pt_enc_get_offset(encoder, &sync[0]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+	errcode = pt_enc_get_offset(encoder, &sync[1]);
+	ptu_int_ge(errcode, 0);
+
+	pt_encode_psb(encoder);
+	pt_encode_mode_exec(encoder, ptem_64bit);
+	pt_encode_psbend(encoder);
+
+
+	errcode = pt_qry_sync_forward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[0]);
+
+	errcode = pt_qry_event(decoder, &event, sizeof(event));
+	ptu_int_ge(errcode, 0);
+	ptu_int_eq(event.type, ptev_exec_mode);
+
+	errcode = pt_qry_event(decoder, &event, sizeof(event));
+	ptu_int_ge(errcode, 0);
+	ptu_int_eq(event.type, ptev_exec_mode);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[1]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_ge(errcode, 0);
+
+	errcode = pt_qry_get_sync_offset(decoder, &offset);
+	ptu_int_eq(errcode, 0);
+	ptu_uint_eq(offset, sync[0]);
+
+	errcode = pt_qry_sync_backward(decoder, &ip);
+	ptu_int_eq(errcode, -pte_eos);
+
+	return ptu_passed();
+}
+
 static struct ptunit_result indir_null(struct ptu_decoder_fixture *dfix)
 {
 	struct pt_query_decoder *decoder = &dfix->decoder;
@@ -803,7 +1104,7 @@ event_async_disabled_suppressed_fail(struct ptu_decoder_fixture *dfix)
 	ptu_check(ptu_sync_decoder, decoder);
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
-	ptu_int_eq(errcode, -pte_bad_packet);
+	ptu_int_eq(errcode, -pte_ip_suppressed);
 
 	return ptu_passed();
 }
@@ -815,11 +1116,9 @@ event_async_disabled_cutoff_fail_a(struct ptu_decoder_fixture *dfix)
 	struct pt_encoder *encoder = &dfix->encoder;
 	struct pt_event event;
 	uint64_t at = pt_dfix_sext_ip;
-	const uint8_t *pos;
 	int errcode;
 
 	pt_encode_fup(encoder, at, pt_ipc_sext_48);
-	pos = encoder->pos;
 	pt_encode_tip_pgd(encoder, 0, pt_ipc_update_16);
 
 	ptu_check(cutoff, decoder, encoder);
@@ -827,7 +1126,6 @@ event_async_disabled_cutoff_fail_a(struct ptu_decoder_fixture *dfix)
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	ptu_int_eq(errcode, -pte_eos);
-	ptu_ptr_eq(decoder->pos, pos);
 
 	return ptu_passed();
 }
@@ -838,10 +1136,8 @@ event_async_disabled_cutoff_fail_b(struct ptu_decoder_fixture *dfix)
 	struct pt_query_decoder *decoder = &dfix->decoder;
 	struct pt_encoder *encoder = &dfix->encoder;
 	struct pt_event event;
-	const uint8_t *pos;
 	int errcode;
 
-	pos = encoder->pos;
 	pt_encode_fup(encoder, 0, pt_ipc_sext_48);
 
 	ptu_check(cutoff, decoder, encoder);
@@ -849,7 +1145,6 @@ event_async_disabled_cutoff_fail_b(struct ptu_decoder_fixture *dfix)
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	ptu_int_eq(errcode, -pte_eos);
-	ptu_ptr_eq(decoder->pos, pos);
 
 	return ptu_passed();
 }
@@ -860,18 +1155,14 @@ event_async_branch_suppressed_fail(struct ptu_decoder_fixture *dfix)
 	struct pt_query_decoder *decoder = &dfix->decoder;
 	struct pt_encoder *encoder = &dfix->encoder;
 	struct pt_event event;
-	const uint8_t *pos;
 	int errcode;
 
-	pos = encoder->pos;
 	pt_encode_fup(encoder, 0, pt_ipc_suppressed);
-	pt_encode_tip(encoder, 0, pt_ipc_sext_48);
 
 	ptu_check(ptu_sync_decoder, decoder);
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
-	ptu_int_eq(errcode, -pte_bad_packet);
-	ptu_ptr_eq(decoder->pos, pos);
+	ptu_int_eq(errcode, -pte_ip_suppressed);
 
 	return ptu_passed();
 }
@@ -926,11 +1217,9 @@ event_async_branch_cutoff_fail_a(struct ptu_decoder_fixture *dfix)
 	struct pt_query_decoder *decoder = &dfix->decoder;
 	struct pt_encoder *encoder = &dfix->encoder;
 	struct pt_event event;
-	const uint8_t *pos;
 	int errcode;
 
 	pt_encode_fup(encoder, 0, pt_ipc_sext_48);
-	pos = encoder->pos;
 	pt_encode_tip_pgd(encoder, 0, pt_ipc_update_16);
 
 	ptu_check(cutoff, decoder, encoder);
@@ -938,7 +1227,6 @@ event_async_branch_cutoff_fail_a(struct ptu_decoder_fixture *dfix)
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	ptu_int_eq(errcode, -pte_eos);
-	ptu_ptr_eq(decoder->pos, pos);
 
 	return ptu_passed();
 }
@@ -949,10 +1237,8 @@ event_async_branch_cutoff_fail_b(struct ptu_decoder_fixture *dfix)
 	struct pt_query_decoder *decoder = &dfix->decoder;
 	struct pt_encoder *encoder = &dfix->encoder;
 	struct pt_event event;
-	const uint8_t *pos;
 	int errcode;
 
-	pos = encoder->pos;
 	pt_encode_fup(encoder, 0, pt_ipc_sext_48);
 
 	ptu_check(cutoff, decoder, encoder);
@@ -960,7 +1246,6 @@ event_async_branch_cutoff_fail_b(struct ptu_decoder_fixture *dfix)
 
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	ptu_int_eq(errcode, -pte_eos);
-	ptu_ptr_eq(decoder->pos, pos);
 
 	return ptu_passed();
 }
@@ -1152,7 +1437,7 @@ static struct ptunit_result event_overflow_fup(struct ptu_decoder_fixture *dfix,
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	switch (ipc) {
 	case pt_ipc_suppressed:
-		ptu_int_eq(errcode, -pte_bad_packet);
+		ptu_int_eq(errcode, -pte_noip);
 		break;
 
 	case pt_ipc_update_16:
@@ -1176,26 +1461,6 @@ static struct ptunit_result event_overflow_fup(struct ptu_decoder_fixture *dfix,
 	return ptu_passed();
 }
 
-static struct ptunit_result
-event_overflow_fup_cutoff_fail(struct ptu_decoder_fixture *dfix)
-{
-	struct pt_query_decoder *decoder = &dfix->decoder;
-	struct pt_encoder *encoder = &dfix->encoder;
-	struct pt_event event;
-	int errcode;
-
-	pt_encode_ovf(encoder);
-	pt_encode_fup(encoder, 0, pt_ipc_sext_48);
-
-	ptu_check(cutoff, decoder, encoder);
-	ptu_check(ptu_sync_decoder, decoder);
-
-	errcode = pt_qry_event(decoder, &event, sizeof(event));
-	ptu_int_eq(errcode, -pte_eos);
-
-	return ptu_passed();
-}
-
 static struct ptunit_result
 event_overflow_tip_pge(struct ptu_decoder_fixture *dfix,
 		       enum pt_ip_compression ipc, uint64_t tsc)
@@ -1752,7 +2017,7 @@ static struct ptunit_result sync_ovf_event(struct ptu_decoder_fixture *dfix,
 	errcode = pt_qry_event(decoder, &event, sizeof(event));
 	switch (ipc) {
 	case pt_ipc_suppressed:
-		ptu_int_eq(errcode, -pte_bad_packet);
+		ptu_int_eq(errcode, -pte_noip);
 		return ptu_passed();
 
 	case pt_ipc_update_16:
@@ -1828,8 +2093,8 @@ static struct ptunit_result time(struct ptu_decoder_fixture *dfix)
 
 	exp = 0x11223344556677ull;
 
-	decoder->time.have_tsc = 1;
-	decoder->time.tsc = exp;
+	decoder->last_time.have_tsc = 1;
+	decoder->last_time.tsc = exp;
 
 	errcode = pt_qry_time(decoder, &tsc, NULL, NULL);
 	ptu_int_eq(errcode, 0);
@@ -1874,8 +2139,8 @@ static struct ptunit_result cbr(struct ptu_decoder_fixture *dfix)
 	uint32_t cbr;
 	int errcode;
 
-	decoder->time.have_cbr = 1;
-	decoder->time.cbr = 42;
+	decoder->last_time.have_cbr = 1;
+	decoder->last_time.cbr = 42;
 
 	errcode = pt_qry_core_bus_ratio(decoder, &cbr);
 	ptu_int_eq(errcode, 0);
@@ -1884,6 +2149,71 @@ static struct ptunit_result cbr(struct ptu_decoder_fixture *dfix)
 	return ptu_passed();
 }
 
+/* Test that end-of-stream is indicated correctly when the stream ends with a
+ * partial non-query-relevant packet.
+ */
+static struct ptunit_result indir_cyc_cutoff(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	uint64_t ip;
+	int errcode;
+
+	pt_encode_tip(encoder, 0xa000ull, pt_ipc_full);
+	pt_encode_cyc(encoder, 0xfff);
+
+	ptu_check(cutoff, decoder, encoder);
+	ptu_check(ptu_sync_decoder, decoder);
+
+	errcode = pt_qry_indirect_branch(decoder, &ip);
+	ptu_int_eq(errcode, pts_eos);
+
+	return ptu_passed();
+}
+
+/* Test that end-of-stream is indicated correctly when the stream ends with a
+ * partial non-query-relevant packet.
+ */
+static struct ptunit_result cond_cyc_cutoff(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	int errcode, taken;
+
+	pt_encode_tnt_8(encoder, 0, 1);
+	pt_encode_cyc(encoder, 0xfff);
+
+	ptu_check(cutoff, decoder, encoder);
+	ptu_check(ptu_sync_decoder, decoder);
+
+	errcode = pt_qry_cond_branch(decoder, &taken);
+	ptu_int_eq(errcode, pts_eos);
+
+	return ptu_passed();
+}
+
+/* Test that end-of-stream is indicated correctly when the stream ends with a
+ * partial non-query-relevant packet.
+ */
+static struct ptunit_result event_cyc_cutoff(struct ptu_decoder_fixture *dfix)
+{
+	struct pt_query_decoder *decoder = &dfix->decoder;
+	struct pt_encoder *encoder = &dfix->encoder;
+	struct pt_event event;
+	int errcode;
+
+	pt_encode_tip_pgd(encoder, 0ull, pt_ipc_full);
+	pt_encode_cyc(encoder, 0xffff);
+
+	ptu_check(cutoff, decoder, encoder);
+	ptu_check(ptu_sync_decoder, decoder);
+
+	errcode = pt_qry_event(decoder, &event, sizeof(event));
+	ptu_int_eq(errcode, pts_eos);
+
+	return ptu_passed();
+}
+
 static struct ptunit_result ptu_dfix_init(struct ptu_decoder_fixture *dfix)
 {
 	struct pt_config *config = &dfix->config;
@@ -1944,7 +2274,7 @@ ptu_dfix_header_indir(struct ptu_decoder_fixture *dfix)
 	struct pt_encoder *encoder = &dfix->encoder;
 
 	pt_encode_pad(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0);
 
@@ -1970,15 +2300,13 @@ ptu_dfix_header_indir_psb(struct ptu_decoder_fixture *dfix)
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0);
 	pt_encode_psb(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0);
 	pt_encode_fup(encoder, pt_dfix_sext_ip, pt_ipc_sext_48);
-	pt_encode_mnt(encoder, 0ull);
 	pt_encode_psbend(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
-	pt_encode_mnt(encoder, 0ull);
 
 	/* Synchronize the decoder at the beginning of the buffer. */
 	decoder->pos = decoder->config.begin;
@@ -2000,17 +2328,15 @@ ptu_dfix_header_cond(struct ptu_decoder_fixture *dfix)
 	 * may want to update last-ip, which requires a last-ip, of course.
 	 */
 	pt_encode_pad(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_psb(encoder);
 	pt_encode_tsc(encoder, 0);
 	pt_encode_pad(encoder);
 	pt_encode_fup(encoder, pt_dfix_sext_ip, pt_ipc_sext_48);
-	pt_encode_mnt(encoder, 0ull);
 	pt_encode_psbend(encoder);
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0);
 	pt_encode_pad(encoder);
-	pt_encode_mnt(encoder, 0ull);
 
 	/* Synchronize the decoder at the beginning of the buffer. */
 	decoder->pos = decoder->config.begin;
@@ -2028,7 +2354,7 @@ ptu_dfix_header_event(struct ptu_decoder_fixture *dfix)
 	struct pt_encoder *encoder = &dfix->encoder;
 
 	pt_encode_pad(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0x1000);
 
@@ -2054,12 +2380,12 @@ ptu_dfix_header_event_psb(struct ptu_decoder_fixture *dfix)
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0);
 	pt_encode_psb(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
 	pt_encode_tsc(encoder, 0x1000);
 	pt_encode_fup(encoder, pt_dfix_sext_ip, pt_ipc_sext_48);
 	pt_encode_psbend(encoder);
-	pt_encode_cbr(encoder, 1);
+	pt_encode_mtc(encoder, 1);
 	pt_encode_pad(encoder);
 
 	/* Synchronize the decoder at the beginning of the buffer. */
@@ -2112,6 +2438,12 @@ int main(int argc, char **argv)
 	ptu_run_f(suite, cond_not_synced, dfix_raw);
 	ptu_run_f(suite, event_not_synced, dfix_raw);
 
+	ptu_run_f(suite, sync_backward, dfix_raw);
+	ptu_run_f(suite, sync_backward_empty_end, dfix_raw);
+	ptu_run_f(suite, sync_backward_empty_mid, dfix_raw);
+	ptu_run_f(suite, sync_backward_empty_begin, dfix_raw);
+	ptu_run_f(suite, decode_sync_backward, dfix_raw);
+
 	ptu_run_f(suite, indir_null, dfix_empty);
 	ptu_run_f(suite, indir_empty, dfix_empty);
 	ptu_run_fp(suite, indir, dfix_empty, pt_ipc_suppressed);
@@ -2237,7 +2569,6 @@ int main(int argc, char **argv)
 	ptu_run_fp(suite, event_overflow_fup, dfix_empty, pt_ipc_update_48, 0);
 	ptu_run_fp(suite, event_overflow_fup, dfix_empty, pt_ipc_sext_48, 0);
 	ptu_run_fp(suite, event_overflow_fup, dfix_empty, pt_ipc_full, 0);
-	ptu_run_f(suite, event_overflow_fup_cutoff_fail, dfix_empty);
 	ptu_run_fp(suite, event_overflow_tip_pge, dfix_empty,
 		   pt_ipc_suppressed, 0);
 	ptu_run_fp(suite, event_overflow_tip_pge, dfix_empty, pt_ipc_update_16,
@@ -2370,7 +2701,6 @@ int main(int argc, char **argv)
 		   0x1000);
 	ptu_run_fp(suite, event_overflow_fup, dfix_event, pt_ipc_full,
 		   0x1000);
-	ptu_run_f(suite, event_overflow_fup_cutoff_fail, dfix_event);
 	ptu_run_fp(suite, event_overflow_tip_pge, dfix_event,
 		   pt_ipc_suppressed, 0x1000);
 	ptu_run_fp(suite, event_overflow_tip_pge, dfix_event, pt_ipc_update_16,
@@ -2535,6 +2865,9 @@ int main(int argc, char **argv)
 	ptu_run_f(suite, cbr_initial, dfix_empty);
 	ptu_run_f(suite, cbr, dfix_empty);
 
-	ptunit_report(&suite);
-	return suite.nr_fails;
+	ptu_run_f(suite, indir_cyc_cutoff, dfix_empty);
+	ptu_run_f(suite, cond_cyc_cutoff, dfix_empty);
+	ptu_run_f(suite, event_cyc_cutoff, dfix_empty);
+
+	return ptunit_report(&suite);
 }
