@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2016, Intel Corporation
+ * Copyright (c) 2013-2018, Intel Corporation
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -29,18 +29,33 @@
 #include "pt_cpu.h"
 #include "pt_last_ip.h"
 #include "pt_time.h"
+#include "pt_compiler.h"
+#include "pt_version.h"
 
 #include "intel-pt.h"
 
+#if defined(FEATURE_SIDEBAND)
+#  include "libipt-sb.h"
+#endif
+
 #include <stdlib.h>
 #include <stdarg.h>
 #include <inttypes.h>
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
+#include <limits.h>
+
+#if defined(_MSC_VER) && (_MSC_VER < 1900)
+#  define snprintf _snprintf_c
+#endif
 
 
 struct ptdump_options {
+#if defined(FEATURE_SIDEBAND)
+	/* Sideband dump flags. */
+	uint32_t sb_dump_flags;
+#endif
 	/* Show the current offset in the trace stream. */
 	uint32_t show_offset:1;
 
@@ -85,6 +100,11 @@ struct ptdump_options {
 
 	/* Don't show CYC packets and ignore them when tracking time. */
 	uint32_t no_cyc:1;
+
+#if defined(FEATURE_SIDEBAND)
+	/* Print sideband warnings. */
+	uint32_t print_sb_warnings:1;
+#endif
 };
 
 struct ptdump_buffer {
@@ -128,6 +148,11 @@ struct ptdump_buffer {
 };
 
 struct ptdump_tracking {
+#if defined(FEATURE_SIDEBAND)
+	/* The sideband session. */
+	struct pt_sb_session *session;
+#endif
+
 	/* Track last-ip. */
 	struct pt_last_ip last_ip;
 
@@ -169,57 +194,74 @@ static int unknown_option_error(const char *arg, const char *name)
 
 static int help(const char *name)
 {
-	fprintf(stderr,
-		"usage: %s [<options>] <ptfile>[:<from>[-<to>]\n\n"
-		"options:\n"
-		"  --help|-h                 this text.\n"
-		"  --version                 display version information and exit.\n"
-		"  --no-sync                 don't try to sync to the first PSB, assume a valid\n"
-		"                            sync point at the beginning of the trace.\n"
-		"  --quiet                   don't print anything but errors.\n"
-		"  --no-pad                  don't show PAD packets.\n"
-		"  --no-timing               don't show timing packets.\n"
-		"  --no-cyc                  don't show CYC packets and ignore them when tracking time.\n"
-		"  --no-offset               don't show the offset as the first column.\n"
-		"  --raw                     show raw packet bytes.\n"
-		"  --lastip                  show last IP updates on packets with IP payloads.\n"
-		"  --exec-mode               show the current execution mode on mode.exec packets.\n"
-		"  --time                    show the estimated TSC on timing packets.\n"
-		"  --tcal                    show time calibration information.\n"
-		"  --time-delta              show timing information as delta.\n"
-		"  --no-tcal                 skip timing calibration.\n"
-		"                            this will result in errors when CYC packets are encountered.\n"
-		"  --no-wall-clock           suppress the no-time error and print relative time.\n"
-		"  --cpu none|auto|f/m[/s]   set cpu to the given value and decode according to:\n"
-		"                              none     spec (default)\n"
-		"                              auto     current cpu\n"
-		"                              f/m[/s]  family/model[/stepping]\n"
-		"  --mtc-freq <n>            set the MTC frequency (IA32_RTIT_CTL[17:14]) to <n>.\n"
-		"  --nom-freq <n>            set the nominal frequency (MSR_PLATFORM_INFO[15:8]) to <n>.\n"
-		"  --cpuid-0x15.eax          set the value of cpuid[0x15].eax.\n"
-		"  --cpuid-0x15.ebx          set the value of cpuid[0x15].ebx.\n"
-		"  <ptfile>[:<from>[-<to>]]  load the processor trace data from <ptfile>;\n"
-		"                            an optional offset or range can be given.\n",
-		name);
+	printf("usage: %s [<options>] <ptfile>[:<from>[-<to>]\n\n", name);
+	printf("options:\n");
+	printf("  --help|-h                 this text.\n");
+	printf("  --version                 display version information and exit.\n");
+	printf("  --no-sync                 don't try to sync to the first PSB, assume a valid\n");
+	printf("                            sync point at the beginning of the trace.\n");
+	printf("  --quiet                   don't print anything but errors.\n");
+	printf("  --no-pad                  don't show PAD packets.\n");
+	printf("  --no-timing               don't show timing packets.\n");
+	printf("  --no-cyc                  don't show CYC packets and ignore them when tracking time.\n");
+	printf("  --no-offset               don't show the offset as the first column.\n");
+	printf("  --raw                     show raw packet bytes.\n");
+	printf("  --lastip                  show last IP updates on packets with IP payloads.\n");
+	printf("  --exec-mode               show the current execution mode on mode.exec packets.\n");
+	printf("  --time                    show the estimated TSC on timing packets.\n");
+	printf("  --tcal                    show time calibration information.\n");
+	printf("  --time-delta              show timing information as delta.\n");
+	printf("  --no-tcal                 skip timing calibration.\n");
+	printf("                            this will result in errors when CYC packets are encountered.\n");
+	printf("  --no-wall-clock           suppress the no-time error and print relative time.\n");
+#if defined(FEATURE_SIDEBAND)
+	printf("  --sb:compact | --sb       show sideband records in compact format.\n");
+	printf("  --sb:verbose              show sideband records in verbose format.\n");
+	printf("  --sb:filename             show the filename on sideband records.\n");
+	printf("  --sb:offset               show the offset on sideband records.\n");
+	printf("  --sb:time                 show the time on sideband records.\n");
+	printf("  --sb:warn                 show sideband warnings.\n");
+#if defined(FEATURE_PEVENT)
+	printf("  --pevent[:primary/:secondary] <file>[:<from>[-<to>]]\n");
+	printf("                              load a perf_event sideband stream from <file>.\n");
+	printf("                              an optional offset or range can be given.\n");
+	printf("  --pevent:sample-type <val>  set perf_event_attr.sample_type to <val> (default: 0).\n");
+	printf("  --pevent:time-zero <val>    set perf_event_mmap_page.time_zero to <val> (default: 0).\n");
+	printf("  --pevent:time-shift <val>   set perf_event_mmap_page.time_shift to <val> (default: 0).\n");
+	printf("  --pevent:time-mult <val>    set perf_event_mmap_page.time_mult to <val> (default: 1).\n");
+	printf("  --pevent:tsc-offset <val>   show perf events <val> ticks earlier.\n");
+	printf("  --pevent:kernel-start <val> the start address of the kernel.\n");
+	printf("  --pevent:sysroot <path>     ignored.\n");
+	printf("  --pevent:kcore <file>       ignored.\n");
+	printf("  --pevent:vdso-x64 <file>    ignored.\n");
+	printf("  --pevent:vdso-x32 <file>    ignored.\n");
+	printf("  --pevent:vdso-ia32 <file>   ignored.\n");
+#endif /* defined(FEATURE_PEVENT) */
+#endif /* defined(FEATURE_SIDEBAND) */
+	printf("  --cpu none|auto|f/m[/s]   set cpu to the given value and decode according to:\n");
+	printf("                              none     spec (default)\n");
+	printf("                              auto     current cpu\n");
+	printf("                              f/m[/s]  family/model[/stepping]\n");
+	printf("  --mtc-freq <n>            set the MTC frequency (IA32_RTIT_CTL[17:14]) to <n>.\n");
+	printf("  --nom-freq <n>            set the nominal frequency (MSR_PLATFORM_INFO[15:8]) to <n>.\n");
+	printf("  --cpuid-0x15.eax          set the value of cpuid[0x15].eax.\n");
+	printf("  --cpuid-0x15.ebx          set the value of cpuid[0x15].ebx.\n");
+	printf("  <ptfile>[:<from>[-<to>]]  load the processor trace data from <ptfile>;\n");
 
-	return 0;
+	return 1;
 }
 
 static int version(const char *name)
 {
-	struct pt_version v = pt_library_version();
-
-	printf("%s-%d.%d.%d%s / libipt-%" PRIu8 ".%" PRIu8 ".%" PRIu32 "%s\n",
-	       name, PT_VERSION_MAJOR, PT_VERSION_MINOR, PT_VERSION_BUILD,
-	       PT_VERSION_EXT, v.major, v.minor, v.build, v.ext);
-	return 0;
+	pt_print_tool_version(name);
+	return 1;
 }
 
-static int parse_range(char *arg, uint64_t *begin, uint64_t *end)
+static int parse_range(const char *arg, uint64_t *begin, uint64_t *end)
 {
 	char *rest;
 
-	if (!arg)
+	if (!arg || !*arg)
 		return 0;
 
 	errno = 0;
@@ -228,7 +270,7 @@ static int parse_range(char *arg, uint64_t *begin, uint64_t *end)
 		return -1;
 
 	if (!*rest)
-		return 0;
+		return 1;
 
 	if (*rest != '-')
 		return -1;
@@ -237,112 +279,162 @@ static int parse_range(char *arg, uint64_t *begin, uint64_t *end)
 	if (errno || *rest)
 		return -1;
 
-	return 0;
+	return 2;
 }
 
-static int load_file(uint8_t **buffer, size_t *size, char *arg,
-		     const char *prog)
+/* Preprocess a filename argument.
+ *
+ * A filename may optionally be followed by a file offset or a file range
+ * argument separated by ':'.  Split the original argument into the filename
+ * part and the offset/range part.
+ *
+ * If no end address is specified, set @size to zero.
+ * If no offset is specified, set @offset to zero.
+ *
+ * Returns zero on success, a negative error code otherwise.
+ */
+static int preprocess_filename(char *filename, uint64_t *offset, uint64_t *size)
+{
+	uint64_t begin, end;
+	char *range;
+	int parts;
+
+	if (!filename || !offset || !size)
+		return -pte_internal;
+
+	/* Search from the end as the filename may also contain ':'. */
+	range = strrchr(filename, ':');
+	if (!range) {
+		*offset = 0ull;
+		*size = 0ull;
+
+		return 0;
+	}
+
+	/* Let's try to parse an optional range suffix.
+	 *
+	 * If we can, remove it from the filename argument.
+	 * If we can not, assume that the ':' is part of the filename, e.g. a
+	 * drive letter on Windows.
+	 */
+	parts = parse_range(range + 1, &begin, &end);
+	if (parts <= 0) {
+		*offset = 0ull;
+		*size = 0ull;
+
+		return 0;
+	}
+
+	if (parts == 1) {
+		*offset = begin;
+		*size = 0ull;
+
+		*range = 0;
+
+		return 0;
+	}
+
+	if (parts == 2) {
+		if (end <= begin)
+			return -pte_invalid;
+
+		*offset = begin;
+		*size = end - begin;
+
+		*range = 0;
+
+		return 0;
+	}
+
+	return -pte_internal;
+}
+
+static int load_file(uint8_t **buffer, size_t *psize, const char *filename,
+		     uint64_t offset, uint64_t size, const char *prog)
 {
-	uint64_t begin_arg, end_arg;
 	uint8_t *content;
 	size_t read;
 	FILE *file;
 	long fsize, begin, end;
 	int errcode;
-	char *range;
 
-	if (!buffer || !size || !arg || !prog) {
+	if (!buffer || !psize || !filename || !prog) {
 		fprintf(stderr, "%s: internal error.\n", prog ? prog : "");
 		return -1;
 	}
 
-	range = strstr(arg, ":");
-	if (range) {
-		range += 1;
-		range[-1] = 0;
-	}
-
 	errno = 0;
-	file = fopen(arg, "rb");
+	file = fopen(filename, "rb");
 	if (!file) {
 		fprintf(stderr, "%s: failed to open %s: %d.\n",
-			prog, arg, errno);
+			prog, filename, errno);
 		return -1;
 	}
 
 	errcode = fseek(file, 0, SEEK_END);
 	if (errcode) {
 		fprintf(stderr, "%s: failed to determine size of %s: %d.\n",
-			prog, arg, errno);
+			prog, filename, errno);
 		goto err_file;
 	}
 
 	fsize = ftell(file);
 	if (fsize < 0) {
 		fprintf(stderr, "%s: failed to determine size of %s: %d.\n",
-			prog, arg, errno);
+			prog, filename, errno);
 		goto err_file;
 	}
 
-	begin_arg = 0ull;
-	end_arg = (uint64_t) fsize;
-	errcode = parse_range(range, &begin_arg, &end_arg);
-	if (errcode < 0) {
-		fprintf(stderr, "%s: bad range: %s.\n", prog, range);
+	begin = (long) offset;
+	if (((uint64_t) begin != offset) || (fsize <= begin)) {
+		fprintf(stderr,
+			"%s: bad offset 0x%" PRIx64 " into %s.\n",
+			prog, offset, filename);
 		goto err_file;
 	}
 
-	begin = (long) begin_arg;
-	end = (long) end_arg;
-	if ((uint64_t) begin != begin_arg || (uint64_t) end != end_arg) {
-		fprintf(stderr, "%s: invalid offset/range argument.\n", prog);
-		goto err_file;
-	}
+	end = fsize;
+	if (size) {
+		uint64_t range_end;
 
-	if (fsize <= begin) {
-		fprintf(stderr, "%s: offset 0x%lx outside of %s.\n",
-			prog, begin, arg);
-		goto err_file;
-	}
+		range_end = offset + size;
+		if ((uint64_t) end < range_end) {
+			fprintf(stderr,
+				"%s: bad range 0x%" PRIx64 " in %s.\n",
+				prog, range_end, filename);
+			goto err_file;
+		}
 
-	if (fsize < end) {
-		fprintf(stderr, "%s: range 0x%lx outside of %s.\n",
-			prog, end, arg);
-		goto err_file;
-	}
-
-	if (end <= begin) {
-		fprintf(stderr, "%s: bad range.\n", prog);
-		goto err_file;
+		end = (long) range_end;
 	}
 
 	fsize = end - begin;
 
-	content = malloc(fsize);
+	content = malloc((size_t) fsize);
 	if (!content) {
 		fprintf(stderr, "%s: failed to allocated memory %s.\n",
-			prog, arg);
+			prog, filename);
 		goto err_file;
 	}
 
 	errcode = fseek(file, begin, SEEK_SET);
 	if (errcode) {
 		fprintf(stderr, "%s: failed to load %s: %d.\n",
-			prog, arg, errno);
+			prog, filename, errno);
 		goto err_content;
 	}
 
-	read = fread(content, fsize, 1, file);
+	read = fread(content, (size_t) fsize, 1u, file);
 	if (read != 1) {
 		fprintf(stderr, "%s: failed to load %s: %d.\n",
-			prog, arg, errno);
+			prog, filename, errno);
 		goto err_content;
 	}
 
 	fclose(file);
 
 	*buffer = content;
-	*size = fsize;
+	*psize = (size_t) fsize;
 
 	return 0;
 
@@ -354,13 +446,14 @@ err_file:
 	return -1;
 }
 
-static int load_pt(struct pt_config *config, char *arg, const char *prog)
+static int load_pt(struct pt_config *config, const char *filename,
+		   uint64_t foffset, uint64_t fsize, const char *prog)
 {
 	uint8_t *buffer;
 	size_t size;
 	int errcode;
 
-	errcode = load_file(&buffer, &size, arg, prog);
+	errcode = load_file(&buffer, &size, filename, foffset, fsize, prog);
 	if (errcode < 0)
 		return errcode;
 
@@ -390,6 +483,9 @@ static void ptdump_tracking_init(struct ptdump_tracking *tracking)
 	pt_tcal_init(&tracking->tcal);
 	pt_time_init(&tracking->time);
 
+#if defined(FEATURE_SIDEBAND)
+	tracking->session = NULL;
+#endif
 	tracking->tsc = 0ull;
 	tracking->fcr = 0ull;
 	tracking->in_header = 0;
@@ -411,9 +507,12 @@ static void ptdump_tracking_reset(struct ptdump_tracking *tracking)
 
 static void ptdump_tracking_fini(struct ptdump_tracking *tracking)
 {
-	(void) tracking;
+	if (!tracking)
+		return;
 
-	/* Nothing to do. */
+#if defined(FEATURE_SIDEBAND)
+	pt_sb_free(tracking->session);
+#endif
 }
 
 #define print_field(field, ...)					\
@@ -566,7 +665,7 @@ static int print_time(struct ptdump_buffer *buffer,
 			if (options->no_wall_clock)
 				break;
 
-			/* Fall through. */
+			fallthrough;
 		default:
 			diag("error printing time", offset, errcode);
 			print_field(buffer->tracking.payload, "<unavailable>");
@@ -639,8 +738,30 @@ static int print_tcal(struct ptdump_buffer *buffer,
 	return 0;
 }
 
+static int sb_track_time(struct ptdump_tracking *tracking,
+			 const struct ptdump_options *options, uint64_t offset)
+{
+	uint64_t tsc;
+	int errcode;
+
+	if (!tracking || !options)
+		return diag("time tracking error", offset, -pte_internal);
+
+	errcode = pt_time_query_tsc(&tsc, NULL, NULL, &tracking->time);
+	if ((errcode < 0) && (errcode != -pte_no_time))
+		return diag("time tracking error", offset, errcode);
+
+#if defined(FEATURE_SIDEBAND)
+	errcode = pt_sb_dump(tracking->session, stdout, options->sb_dump_flags,
+			     tsc);
+	if (errcode < 0)
+		return diag("sideband dump error", offset, errcode);
+#endif
+	return 0;
+}
+
 static int track_time(struct ptdump_buffer *buffer,
-		      struct ptdump_tracking *tracking,  uint64_t offset,
+		      struct ptdump_tracking *tracking, uint64_t offset,
 		      const struct ptdump_options *options)
 {
 	if (!tracking || !options)
@@ -652,7 +773,7 @@ static int track_time(struct ptdump_buffer *buffer,
 	if (options->show_time && !buffer->skip_time)
 		print_time(buffer, tracking, offset, options);
 
-	return 0;
+	return sb_track_time(tracking, options, offset);
 }
 
 static int track_tsc(struct ptdump_buffer *buffer,
@@ -903,6 +1024,35 @@ static const char *print_exec_mode(const struct pt_packet_mode_exec *packet,
 	return "invalid";
 }
 
+static const char *print_pwrx_wr(const struct pt_packet_pwrx *packet)
+{
+	const char *wr;
+
+	if (!packet)
+		return "err";
+
+	wr = NULL;
+	if (packet->interrupt)
+		wr = "int";
+
+	if (packet->store) {
+		if (wr)
+			return NULL;
+		wr = " st";
+	}
+
+	if (packet->autonomous) {
+		if (wr)
+			return NULL;
+		wr = " hw";
+	}
+
+	if (!wr)
+		wr = "bad";
+
+	return wr;
+}
+
 static int print_packet(struct ptdump_buffer *buffer, uint64_t offset,
 			const struct pt_packet *packet,
 			struct ptdump_tracking *tracking,
@@ -1136,6 +1286,50 @@ static int print_packet(struct ptdump_buffer *buffer, uint64_t offset,
 		print_field(buffer->payload.standard, "%" PRIx64,
 			    packet->payload.mnt.payload);
 		return 0;
+
+	case ppt_exstop:
+		print_field(buffer->opcode, "exstop");
+		print_field(buffer->payload.standard, "%s",
+			    packet->payload.exstop.ip ? "ip" : "");
+		return 0;
+
+	case ppt_mwait:
+		print_field(buffer->opcode, "mwait");
+		print_field(buffer->payload.standard, "%08x, %08x",
+			    packet->payload.mwait.hints,
+			    packet->payload.mwait.ext);
+		return 0;
+
+	case ppt_pwre:
+		print_field(buffer->opcode, "pwre");
+		print_field(buffer->payload.standard, "c%u.%u%s",
+			    (packet->payload.pwre.state + 1) & 0xf,
+			    (packet->payload.pwre.sub_state + 1) & 0xf,
+			    packet->payload.pwre.hw ? ", hw" : "");
+		return 0;
+
+	case ppt_pwrx: {
+		const char *wr;
+
+		wr = print_pwrx_wr(&packet->payload.pwrx);
+		if (!wr)
+			wr = "bad";
+
+		print_field(buffer->opcode, "pwrx");
+		print_field(buffer->payload.standard, "%s: c%u, c%u", wr,
+			    (packet->payload.pwrx.last + 1) & 0xf,
+			    (packet->payload.pwrx.deepest + 1) & 0xf);
+		return 0;
+	}
+
+	case ppt_ptw:
+		print_field(buffer->opcode, "ptw");
+		print_field(buffer->payload.standard, "%x: %" PRIx64 "%s",
+			    packet->payload.ptw.plc,
+			    packet->payload.ptw.payload,
+			    packet->payload.ptw.ip ? ", ip" : "");
+
+		return 0;
 	}
 
 	return diag("unknown packet", offset, -pte_bad_opc);
@@ -1214,8 +1408,12 @@ static int dump_sync(struct pt_packet_decoder *decoder,
 			return diag("sync error", 0ull, errcode);
 	} else {
 		errcode = pt_pkt_sync_forward(decoder);
-		if (errcode < 0)
+		if (errcode < 0) {
+			if (errcode == -pte_eos)
+				return 0;
+
 			return diag("sync error", 0ull, errcode);
+		}
 	}
 
 	for (;;) {
@@ -1224,8 +1422,12 @@ static int dump_sync(struct pt_packet_decoder *decoder,
 			break;
 
 		errcode = pt_pkt_sync_forward(decoder);
-		if (errcode < 0)
+		if (errcode < 0) {
+			if (errcode == -pte_eos)
+				return 0;
+
 			return diag("sync error", 0ull, errcode);
+		}
 
 		ptdump_tracking_reset(tracking);
 	}
@@ -1233,26 +1435,123 @@ static int dump_sync(struct pt_packet_decoder *decoder,
 	return errcode;
 }
 
-static int dump(const struct pt_config *config,
+static int dump(struct ptdump_tracking *tracking,
+		const struct pt_config *config,
 		const struct ptdump_options *options)
 {
 	struct pt_packet_decoder *decoder;
-	struct ptdump_tracking tracking;
 	int errcode;
 
 	decoder = pt_pkt_alloc_decoder(config);
 	if (!decoder)
 		return diag("failed to allocate decoder", 0ull, 0);
 
-	ptdump_tracking_init(&tracking);
+	errcode = dump_sync(decoder, tracking, options, config);
 
-	errcode = dump_sync(decoder, &tracking, options, config);
-
-	ptdump_tracking_fini(&tracking);
 	pt_pkt_free_decoder(decoder);
-	return errcode;
+
+	if (errcode < 0)
+		return errcode;
+
+#if defined(FEATURE_SIDEBAND)
+	errcode = pt_sb_dump(tracking->session, stdout, options->sb_dump_flags,
+			     UINT64_MAX);
+	if (errcode < 0)
+		return diag("sideband dump error", UINT64_MAX, errcode);
+#endif
+
+	return 0;
 }
 
+#if defined(FEATURE_SIDEBAND)
+
+static int ptdump_print_error(int errcode, const char *filename,
+			      uint64_t offset, void *priv)
+{
+	const struct ptdump_options *options;
+	const char *errstr;
+
+	options = (struct ptdump_options *) priv;
+	if (!options)
+		return -pte_internal;
+
+	if (errcode >= 0 && !options->print_sb_warnings)
+		return 0;
+
+	if (!filename)
+		filename = "<unknown>";
+
+	errstr = errcode < 0
+		? pt_errstr(pt_errcode(errcode))
+		: pt_sb_errstr((enum pt_sb_error_code) errcode);
+
+	if (!errstr)
+		errstr = "<unknown error>";
+
+	printf("[%s:%016" PRIx64 " sideband error: %s]\n", filename, offset,
+	       errstr);
+
+	return 0;
+}
+
+#if defined(FEATURE_PEVENT)
+
+static int ptdump_sb_pevent(struct pt_sb_session *session, char *filename,
+			    const struct pt_sb_pevent_config *conf,
+			    const char *prog)
+{
+	struct pt_sb_pevent_config config;
+	uint64_t foffset, fsize, fend;
+	int errcode;
+
+	if (!conf || !prog) {
+		fprintf(stderr, "%s: internal error.\n", prog ? prog : "");
+		return -1;
+	}
+
+	errcode = preprocess_filename(filename, &foffset, &fsize);
+	if (errcode < 0) {
+		fprintf(stderr, "%s: bad file %s: %s.\n", prog, filename,
+			pt_errstr(pt_errcode(errcode)));
+		return -1;
+	}
+
+	if (SIZE_MAX < foffset) {
+		fprintf(stderr,
+			"%s: bad offset: 0x%" PRIx64 ".\n", prog, foffset);
+		return -1;
+	}
+
+	config = *conf;
+	config.filename = filename;
+	config.begin = (size_t) foffset;
+	config.end = 0;
+
+	if (fsize) {
+		fend = foffset + fsize;
+		if ((fend <= foffset) || (SIZE_MAX < fend)) {
+			fprintf(stderr,
+				"%s: bad range: 0x%" PRIx64 "-0x%" PRIx64 ".\n",
+				prog, foffset, fend);
+			return -1;
+		}
+
+		config.end = (size_t) fend;
+	}
+
+	errcode = pt_sb_alloc_pevent_decoder(session, &config);
+	if (errcode < 0) {
+		fprintf(stderr, "%s: error loading %s: %s.\n", prog, filename,
+			pt_errstr(pt_errcode(errcode)));
+		return -1;
+	}
+
+	return 0;
+}
+
+#endif /* defined(FEATURE_PEVENT) */
+#endif /* defined(FEATURE_SIDEBAND) */
+
 static int get_arg_uint64(uint64_t *value, const char *option, const char *arg,
 			  const char *prog)
 {
@@ -1263,7 +1562,7 @@ static int get_arg_uint64(uint64_t *value, const char *option, const char *arg,
 		return 0;
 	}
 
-	if (!arg || (arg[0] == '-' && arg[1] == '-')) {
+	if (!arg || arg[0] == 0 || (arg[0] == '-' && arg[1] == '-')) {
 		fprintf(stderr, "%s: %s: missing argument.\n", prog, option);
 		return 0;
 	}
@@ -1298,6 +1597,29 @@ static int get_arg_uint32(uint32_t *value, const char *option, const char *arg,
 	return 1;
 }
 
+#if defined(FEATURE_SIDEBAND) && defined(FEATURE_PEVENT)
+
+static int get_arg_uint16(uint16_t *value, const char *option, const char *arg,
+			  const char *prog)
+{
+	uint64_t val;
+
+	if (!get_arg_uint64(&val, option, arg, prog))
+		return 0;
+
+	if (val > UINT16_MAX) {
+		fprintf(stderr, "%s: %s: value too big: %s.\n", prog, option,
+			arg);
+		return 0;
+	}
+
+	*value = (uint16_t) val;
+
+	return 1;
+}
+
+#endif /* defined(FEATURE_SIDEBAND) && defined(FEATURE_PEVENT) */
+
 static int get_arg_uint8(uint8_t *value, const char *option, const char *arg,
 			 const char *prog)
 {
@@ -1317,24 +1639,29 @@ static int get_arg_uint8(uint8_t *value, const char *option, const char *arg,
 	return 1;
 }
 
-int main(int argc, char *argv[])
+static int process_args(int argc, char *argv[],
+			struct ptdump_tracking *tracking,
+			struct ptdump_options *options,
+			struct pt_config *config, char **ptfile)
 {
-	struct ptdump_options options;
-	struct pt_config config;
-	int errcode, idx;
-	char *ptfile;
+#if defined(FEATURE_SIDEBAND) && defined(FEATURE_PEVENT)
+	struct pt_sb_pevent_config pevent;
+#endif
+	int idx, errcode;
 
-	ptfile = NULL;
-
-	memset(&options, 0, sizeof(options));
-	options.show_offset = 1;
-
-	memset(&config, 0, sizeof(config));
-	pt_config_init(&config);
+	if (!argv || !tracking || !options || !config || !ptfile) {
+		fprintf(stderr, "%s: internal error.\n", argv ? argv[0] : "");
+		return -1;
+	}
 
+#if defined(FEATURE_SIDEBAND) && defined(FEATURE_PEVENT)
+	memset(&pevent, 0, sizeof(pevent));
+	pevent.size = sizeof(pevent);
+	pevent.time_mult = 1;
+#endif
 	for (idx = 1; idx < argc; ++idx) {
 		if (strncmp(argv[idx], "-", 1) != 0) {
-			ptfile = argv[idx];
+			*ptfile = argv[idx];
 			if (idx < (argc-1))
 				return usage(argv[0]);
 			break;
@@ -1347,47 +1674,138 @@ int main(int argc, char *argv[])
 		if (strcmp(argv[idx], "--version") == 0)
 			return version(argv[0]);
 		if (strcmp(argv[idx], "--no-sync") == 0)
-			options.no_sync = 1;
-		else if (strcmp(argv[idx], "--quiet") == 0)
-			options.quiet = 1;
-		else if (strcmp(argv[idx], "--no-pad") == 0)
-			options.no_pad = 1;
+			options->no_sync = 1;
+		else if (strcmp(argv[idx], "--quiet") == 0) {
+			options->quiet = 1;
+#if defined(FEATURE_SIDEBAND)
+			options->sb_dump_flags = 0;
+#endif
+		} else if (strcmp(argv[idx], "--no-pad") == 0)
+			options->no_pad = 1;
 		else if (strcmp(argv[idx], "--no-timing") == 0)
-			options.no_timing = 1;
+			options->no_timing = 1;
 		else if (strcmp(argv[idx], "--no-cyc") == 0)
-			options.no_cyc = 1;
+			options->no_cyc = 1;
 		else if (strcmp(argv[idx], "--no-offset") == 0)
-			options.show_offset = 0;
+			options->show_offset = 0;
 		else if (strcmp(argv[idx], "--raw") == 0)
-			options.show_raw_bytes = 1;
+			options->show_raw_bytes = 1;
 		else if (strcmp(argv[idx], "--lastip") == 0)
-			options.show_last_ip = 1;
+			options->show_last_ip = 1;
 		else if (strcmp(argv[idx], "--exec-mode") == 0)
-			options.show_exec_mode = 1;
+			options->show_exec_mode = 1;
 		else if (strcmp(argv[idx], "--time") == 0) {
-			if (options.show_tcal) {
+			if (options->show_tcal) {
 				fprintf(stderr, "%s: specify either --time "
 					"or --tcal.\n", argv[0]);
-				return 1;
+				return -1;
 			}
 
-			options.track_time = 1;
-			options.show_time = 1;
+			options->track_time = 1;
+			options->show_time = 1;
 		} else if (strcmp(argv[idx], "--time-delta") == 0) {
-			options.show_time_as_delta = 1;
+			options->show_time_as_delta = 1;
 		} else if (strcmp(argv[idx], "--tcal") == 0) {
-			if (options.show_time) {
+			if (options->show_time) {
 				fprintf(stderr, "%s: specify either --time "
 					"or --tcal.\n", argv[0]);
-				return 1;
+				return -1;
 			}
 
-			options.track_time = 1;
-			options.show_tcal = 1;
+			options->track_time = 1;
+			options->show_tcal = 1;
 		} else if (strcmp(argv[idx], "--no-tcal") == 0)
-			options.no_tcal = 1;
+			options->no_tcal = 1;
 		else if (strcmp(argv[idx], "--no-wall-clock") == 0)
-			options.no_wall_clock = 1;
+			options->no_wall_clock = 1;
+#if defined(FEATURE_SIDEBAND)
+		else if ((strcmp(argv[idx], "--sb:compact") == 0) ||
+			 (strcmp(argv[idx], "--sb") == 0)) {
+			options->sb_dump_flags &= ~(uint32_t) ptsbp_verbose;
+			options->sb_dump_flags |= (uint32_t) ptsbp_compact;
+		} else if (strcmp(argv[idx], "--sb:verbose") == 0) {
+			options->sb_dump_flags &= ~(uint32_t) ptsbp_compact;
+			options->sb_dump_flags |= (uint32_t) ptsbp_verbose;
+		} else if (strcmp(argv[idx], "--sb:filename") == 0)
+			options->sb_dump_flags |= (uint32_t) ptsbp_filename;
+		else if (strcmp(argv[idx], "--sb:offset") == 0)
+			options->sb_dump_flags |= (uint32_t) ptsbp_file_offset;
+		else if (strcmp(argv[idx], "--sb:time") == 0)
+			options->sb_dump_flags |= (uint32_t) ptsbp_tsc;
+		else if (strcmp(argv[idx], "--sb:warn") == 0)
+			options->print_sb_warnings = 1;
+#if defined(FEATURE_PEVENT)
+		else if ((strcmp(argv[idx], "--pevent") == 0) ||
+			 (strcmp(argv[idx], "--pevent:primary") == 0) ||
+			 (strcmp(argv[idx], "--pevent:secondary") == 0)) {
+			char *arg;
+
+			arg = argv[++idx];
+			if (!arg) {
+				fprintf(stderr,
+					"%s: %s: missing argument.\n",
+					argv[0], argv[idx-1]);
+				return -1;
+			}
+
+			errcode = ptdump_sb_pevent(tracking->session, arg,
+						   &pevent, argv[0]);
+			if (errcode < 0)
+				return -1;
+
+			/* We need to keep track of time for sideband
+			 * correlation.
+			 */
+			options->track_time = 1;
+		} else if (strcmp(argv[idx], "--pevent:sample-type") == 0) {
+			if (!get_arg_uint64(&pevent.sample_type,
+					    "--pevent:sample-type",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if (strcmp(argv[idx], "--pevent:time-zero") == 0) {
+			if (!get_arg_uint64(&pevent.time_zero,
+					    "--pevent:time-zero",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if (strcmp(argv[idx], "--pevent:time-shift") == 0) {
+			if (!get_arg_uint16(&pevent.time_shift,
+					    "--pevent:time-shift",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if (strcmp(argv[idx], "--pevent:time-mult") == 0) {
+			if (!get_arg_uint32(&pevent.time_mult,
+					    "--pevent:time-mult",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if (strcmp(argv[idx], "--pevent:tsc-offset") == 0) {
+			if (!get_arg_uint64(&pevent.tsc_offset,
+					    "--pevent:tsc-offset",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if (strcmp(argv[idx], "--pevent:kernel-start") == 0) {
+			if (!get_arg_uint64(&pevent.kernel_start,
+					    "--pevent:kernel-start",
+					    argv[++idx], argv[0]))
+				return -1;
+		} else if ((strcmp(argv[idx], "--pevent:sysroot") == 0) ||
+			   (strcmp(argv[idx], "--pevent:kcore") == 0) ||
+			   (strcmp(argv[idx], "--pevent:vdso-x64") == 0) ||
+			   (strcmp(argv[idx], "--pevent:vdso-x32") == 0) ||
+			   (strcmp(argv[idx], "--pevent:vdso-ia32") == 0)) {
+			char *arg;
+
+			arg = argv[++idx];
+			if (!arg) {
+				fprintf(stderr,
+					"%s: %s: missing argument.\n",
+					argv[0], argv[idx-1]);
+				return -1;
+			}
+
+			/* Ignore. */
+		}
+#endif /* defined(FEATURE_PEVENT) */
+#endif /* defined(FEATURE_SIDEBAND) */
 		else if (strcmp(argv[idx], "--cpu") == 0) {
 			const char *arg;
 
@@ -1396,69 +1814,134 @@ int main(int argc, char *argv[])
 				fprintf(stderr,
 					"%s: --cpu: missing argument.\n",
 					argv[0]);
-				return 1;
+				return -1;
 			}
 
 			if (strcmp(arg, "auto") == 0) {
-				errcode = pt_cpu_read(&config.cpu);
+				errcode = pt_cpu_read(&config->cpu);
 				if (errcode < 0) {
 					fprintf(stderr,
 						"%s: error reading cpu: %s.\n",
 						argv[0],
 						pt_errstr(pt_errcode(errcode)));
-					return 1;
+					return -1;
 				}
 				continue;
 			}
 
 			if (strcmp(arg, "none") == 0) {
-				memset(&config.cpu, 0, sizeof(config.cpu));
+				memset(&config->cpu, 0, sizeof(config->cpu));
 				continue;
 			}
 
-			errcode = pt_cpu_parse(&config.cpu, arg);
+			errcode = pt_cpu_parse(&config->cpu, arg);
 			if (errcode < 0) {
 				fprintf(stderr,
 					"%s: cpu must be specified as f/m[/s]\n",
 					argv[0]);
-				return 1;
+				return -1;
 			}
 		} else if (strcmp(argv[idx], "--mtc-freq") == 0) {
-			if (!get_arg_uint8(&config.mtc_freq, "--mtc-freq",
+			if (!get_arg_uint8(&config->mtc_freq, "--mtc-freq",
 					   argv[++idx], argv[0]))
-				return 1;
+				return -1;
 		} else if (strcmp(argv[idx], "--nom-freq") == 0) {
-			if (!get_arg_uint8(&config.nom_freq, "--nom-freq",
+			if (!get_arg_uint8(&config->nom_freq, "--nom-freq",
 					   argv[++idx], argv[0]))
-				return 1;
+				return -1;
 		} else if (strcmp(argv[idx], "--cpuid-0x15.eax") == 0) {
-			if (!get_arg_uint32(&config.cpuid_0x15_eax,
+			if (!get_arg_uint32(&config->cpuid_0x15_eax,
 					    "--cpuid-0x15.eax", argv[++idx],
 					    argv[0]))
-				return 1;
+				return -1;
 		} else if (strcmp(argv[idx], "--cpuid-0x15.ebx") == 0) {
-			if (!get_arg_uint32(&config.cpuid_0x15_ebx,
+			if (!get_arg_uint32(&config->cpuid_0x15_ebx,
 					    "--cpuid-0x15.ebx", argv[++idx],
 					    argv[0]))
-				return 1;
+				return -1;
 		} else
 			return unknown_option_error(argv[idx], argv[0]);
 	}
 
-	if (!ptfile)
-		return no_file_error(argv[0]);
+	return 0;
+}
 
-	errcode = pt_cpu_errata(&config.errata, &config.cpu);
+int main(int argc, char *argv[])
+{
+	struct ptdump_tracking tracking;
+	struct ptdump_options options;
+	struct pt_config config;
+	int errcode;
+	char *ptfile;
+	uint64_t pt_offset, pt_size;
+
+	ptfile = NULL;
+
+	memset(&options, 0, sizeof(options));
+	options.show_offset = 1;
+
+	memset(&config, 0, sizeof(config));
+	pt_config_init(&config);
+
+	ptdump_tracking_init(&tracking);
+
+#if defined(FEATURE_SIDEBAND)
+	tracking.session = pt_sb_alloc(NULL);
+	if (!tracking.session) {
+		fprintf(stderr,
+			"%s: failed to allocate sideband session.\n", argv[0]);
+		errcode = -pte_nomem;
+		goto out;
+	}
+
+	pt_sb_notify_error(tracking.session, ptdump_print_error, &options);
+#endif /* defined(FEATURE_SIDEBAND) */
+
+	errcode = process_args(argc, argv, &tracking, &options, &config,
+			       &ptfile);
+	if (errcode != 0) {
+		if (errcode > 0)
+			errcode = 0;
+		goto out;
+	}
+
+	if (!ptfile) {
+		errcode = no_file_error(argv[0]);
+		goto out;
+	}
+
+	errcode = preprocess_filename(ptfile, &pt_offset, &pt_size);
+	if (errcode < 0) {
+		fprintf(stderr, "%s: bad file %s: %s.\n", argv[0], ptfile,
+			pt_errstr(pt_errcode(errcode)));
+		goto out;
+	}
+
+	if (config.cpu.vendor) {
+		errcode = pt_cpu_errata(&config.errata, &config.cpu);
+		if (errcode < 0)
+			diag("failed to determine errata", 0ull, errcode);
+	}
+
+	errcode = load_pt(&config, ptfile, pt_offset, pt_size, argv[0]);
 	if (errcode < 0)
-		diag("failed to determine errata", 0ull, errcode);
+		goto out;
 
-	errcode = load_pt(&config, ptfile, argv[0]);
-	if (errcode < 0)
-		return errcode;
+#if defined(FEATURE_SIDEBAND)
+	errcode = pt_sb_init_decoders(tracking.session);
+	if (errcode < 0) {
+		fprintf(stderr,
+			"%s: error initializing sideband decoders: %s.\n",
+			argv[0], pt_errstr(pt_errcode(errcode)));
+		goto out;
+	}
+#endif /* defined(FEATURE_SIDEBAND) */
 
-	errcode = dump(&config, &options);
+	errcode = dump(&tracking, &config, &options);
 
+out:
 	free(config.begin);
+	ptdump_tracking_fini(&tracking);
 
 	return -errcode;
 }
