@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2016, Intel Corporation
+ * Copyright (c) 2013-2018, Intel Corporation
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -38,6 +38,10 @@
 #include <stdlib.h>
 #include <string.h>
 
+#if defined(_MSC_VER) && (_MSC_VER < 1900)
+#  define snprintf _snprintf_c
+#endif
+
 
 static int create_section_label_name(char *label, int size, const char *name,
 				     const char *attribute)
@@ -186,7 +190,18 @@ int parse_yasm_labels(struct label *l, const struct text *t)
 
 		tmp = strstr(line, "[org");
 		if (tmp) {
-			base_addr = strtol(tmp+strlen("[org"), NULL, 0);
+			char *org;
+
+			org = tmp + strlen("[org");
+			tmp = strstr(org, "]");
+			if (!tmp)
+				return -err_no_org_directive;
+
+			*tmp = 0;
+
+			errcode = str_to_uint64(org, &base_addr, 0);
+			if (errcode < 0)
+				return errcode;
 
 			errcode = l_append(l, "org", base_addr);
 			if (errcode < 0)
@@ -283,6 +298,8 @@ int parse_yasm_labels(struct label *l, const struct text *t)
 			continue;
 
 		if (!make_label(tmp)) {
+			uint64_t laddr;
+
 			/* get address in case we find a label later.  */
 			if (sscanf(tmp, "%" PRIx64, &addr) != 1)
 				continue;
@@ -297,7 +314,13 @@ int parse_yasm_labels(struct label *l, const struct text *t)
 			if (!make_label(tmp))
 				continue;
 
-			errcode = l_append(l, tmp, addr + base_addr);
+			laddr = addr + base_addr;
+			if (laddr < base_addr) {
+				errcode = -err_label_addr;
+				goto error;
+			}
+
+			errcode = l_append(l, tmp, laddr);
 			if (errcode < 0)
 				goto error;
 			continue;
@@ -322,7 +345,15 @@ int parse_yasm_labels(struct label *l, const struct text *t)
 			}
 			if (sscanf(line, "%*d %" PRIx64 " %*x %*s", &addr)
 			    == 1) {
-				errcode = l_append(l, name, addr + base_addr);
+				uint64_t laddr;
+
+				laddr = addr + base_addr;
+				if (laddr < base_addr) {
+					errcode = -err_label_addr;
+					break;
+				}
+
+				errcode = l_append(l, name, laddr);
 				break;
 			}
 		}
@@ -465,7 +496,8 @@ void pd_free(struct pt_directive *pd)
 	free(pd);
 }
 
-int pd_set(struct pt_directive *pd, const char *name, const char *payload)
+int pd_set(struct pt_directive *pd, enum pt_directive_kind kind,
+	   const char *name, const char *payload)
 {
 	if (bug_on(!pd))
 		return -err_internal;
@@ -476,6 +508,7 @@ int pd_set(struct pt_directive *pd, const char *name, const char *payload)
 	if (bug_on(!payload))
 		return -err_internal;
 
+	pd->kind = kind;
 	strncpy(pd->name, name, pd->nlen);
 	if (pd->nlen > 0)
 		pd->name[pd->nlen - 1] = '\0';
@@ -486,13 +519,17 @@ int pd_set(struct pt_directive *pd, const char *name, const char *payload)
 	return 0;
 }
 
-/* Magic annotation marker.  */
-const char *marker = "@pt ";
+/* Magic annotation markers.  */
+static const char *pt_marker = "@pt ";
+
+#if defined(FEATURE_SIDEBAND)
+static const char *sb_marker = "@sb ";
+#endif
 
 int pd_parse(struct pt_directive *pd, struct state *st)
 {
-	char *line, *comment, *ptdirective, *openpar, *closepar;
-	char *directive, *payload;
+	enum pt_directive_kind kind;
+	char *line, *comment, *openpar, *closepar, *directive, *payload;
 	int errcode;
 	char *c;
 
@@ -526,13 +563,25 @@ int pd_parse(struct pt_directive *pd, struct state *st)
 		goto cleanup;
 
 	/* search for @pt marker.  */
-	ptdirective = strstr(comment+1, marker);
+	directive = strstr(comment+1, pt_marker);
+	if (directive) {
+		directive += strlen(pt_marker);
+		kind = pdk_pt;
+	} else {
+#if defined(FEATURE_SIDEBAND)
+		/* search for @sb marker. */
+		directive = strstr(comment+1, sb_marker);
+		if (directive) {
+			directive += strlen(sb_marker);
+			kind = pdk_sb;
+		} else
+#endif
+			goto cleanup;
+	}
 
-	/* if there is no such marker in the comment, we don't have
-	 * anything to do.
-	 */
-	if (!ptdirective)
-		goto cleanup;
+	/* skip leading whitespace. */
+	while (isspace(*directive))
+		directive += 1;
 
 	/* directive found, now parse the payload.  */
 	errcode = 0;
@@ -540,7 +589,7 @@ int pd_parse(struct pt_directive *pd, struct state *st)
 	/* find position of next '(', separating the directive and the
 	 * payload.
 	 */
-	openpar = strchr(ptdirective, '(');
+	openpar = strchr(directive, '(');
 	if (!openpar) {
 		errcode = -err_missing_openpar;
 		st_print_err(st, "invalid syntax", errcode);
@@ -562,23 +611,18 @@ int pd_parse(struct pt_directive *pd, struct state *st)
 	*openpar = '\0';
 	*closepar = '\0';
 
-	/* skip leading whitespace. */
-	directive = ptdirective + strlen(marker);
-	while (isspace(*directive))
-		directive += 1;
-
 	payload = openpar+1;
 
-	errcode = pd_set(pd, directive, payload);
+	errcode = pd_set(pd, kind, directive, payload);
 
 cleanup:
 	free(line);
 	return errcode;
 }
 
-const char *bin_suffix = ".bin";
-const char *lst_suffix = ".lst";
-const char path_separator = '/';
+static const char *bin_suffix = ".bin";
+static const char *lst_suffix = ".lst";
+static const char path_separator = '/';
 enum {
 	max_filename_len = 1024
 };
@@ -696,6 +740,7 @@ void yasm_free(struct yasm *y)
 	if (!y)
 		return;
 
+	free(y->fileroot);
 	free(y->pttfile);
 	free(y->lstfile);
 	free(y->binfile);
@@ -728,8 +773,8 @@ static int yasm_advance_next_line(struct yasm *y)
 
 
 	for (;;) {
-		errcode = fl_getline(y->fl, s, slen, y->lstfile,
-				     y->lst_curr_line);
+		errcode = fl_getline(y->fl, s, (size_t) slen, y->lstfile,
+				     (size_t) y->lst_curr_line);
 		/* always advance in lst file.  */
 		y->lst_curr_line += 1;
 
@@ -740,8 +785,8 @@ static int yasm_advance_next_line(struct yasm *y)
 		 * state information to this file, line and increment
 		 * and continue.
 		 */
-		if (sscanf(s, "%*d %%line %d+%d %1023s", &asm_line, &asm_inc,
-			   filename) == 3) {
+		if (sscanf(s, "%*d %%line %d+%d %1023[^\r\n]", &asm_line,
+			   &asm_inc, filename) == 3) {
 			st_set_file(y->st_asm, filename, asm_line, asm_inc);
 			continue;
 		}
@@ -757,8 +802,9 @@ static int yasm_advance_next_line(struct yasm *y)
 		 * correlated to the source file, so we retrieve the
 		 * line from it and update the state.
 		 */
-		errcode = fl_getline(y->fl, s, slen, y->st_asm->filename,
-				     y->st_asm->n-1);
+		errcode = fl_getline(y->fl, s, (size_t) slen,
+				     y->st_asm->filename,
+				     (size_t) y->st_asm->n - 1u);
 		st_update(y->st_asm, s);
 		break;
 	}
